﻿
' ==================================================================================================
'   CLASS Module_InOut
' ==================================================================================================

Partial Friend Class Module_InOut

    Friend Adc24_MasterPinName As Int32 = 60
    Friend Adc24_FirstPinName As Int32 = 61
    Friend Adc24_LastPinName As Int32 = 76
    Friend Generic_FirstPinName As Int32 = 80

    Friend Adc24_MasterPinIndex As Int32
    Friend Adc24_FirstChannelIndex As Int32

    Private Const MaxPinCount As Int32 = 255
    Friend OutputDataBuffer(MaxPinCount) As Single
    Friend InputDataBuffer(MaxPinCount) As Single

    Friend ProgModuleCompleted As Boolean = False
    Friend SinglePinPropsRequestCompleted As Boolean = False
    Friend PwmPinDataCompleted As Boolean = False
    Friend StepperPinDataCompleted As Boolean = False
    Friend Adc24ConfigCompleted As Boolean = False
    Friend DataExchangeCompleted As Boolean = True

    ' Y = Programmed OK
    ' N = Not programmed 
    ' W = Waiting for Adc configuration
    ' M = Mismatch between requested and available PinTypes
    Friend ProgrammedStatus As String = ""

    Friend MyDeviceIsActive As Boolean

    Private mActionLock As Object = New Object
    Private mMaxFps As Int32

    Friend ConfigId As Int32
    Friend ConfigValid As Boolean

    Friend ModuleId As Int32
    Friend ModuleVersion As String
    Private ModuleName As String

    Friend CommSpeed As Int32
    Friend CommMillisec As Single
    Friend CommFps As Single
    Friend ErrorRate As Single

    Friend IpAddr As String
    Friend ProgStatus As String

    Private PinsPropsProgrammedByIOTHAL() As NetHALRPC.PinProps
    Private PinsCapsFromIotModule() As String



    ' ==========================================================================================
    '   PUBLIC COMMANDS - COMMUNICATION 
    ' ==========================================================================================
    Friend Function OpenCommunication() As Boolean
        ' ready to open serial ports or similar
        Return True
    End Function

    Friend Sub CloseCommunication()
        ' ready to close serial ports or similar
    End Sub

    Friend Sub SetCommSpeed(ByVal _speed As Int32)
        CommSpeed = _speed
        Select Case CommSpeed
            Case 1 : mMaxFps = 10
            Case 2 : mMaxFps = 20
            Case 3 : mMaxFps = 30
            Case 4 : mMaxFps = 50
            Case 5 : mMaxFps = 60
            Case 6 : mMaxFps = 100
            Case 7 : mMaxFps = 150
            Case 8 : mMaxFps = 200
            Case 9 : mMaxFps = 300
            Case 10 : mMaxFps = 400
            Case 11 : mMaxFps = 500
            Case Else : mMaxFps = 9999
        End Select
    End Sub

    Friend Sub SetName(ByVal _name As String)
        ModuleName = _name
    End Sub

    Friend Function GetName() As String
        Return ModuleName
    End Function

    Friend Sub New(ByVal _moduleid As Int32)
        ModuleName = ""
        ModuleId = _moduleid
        ModuleVersion = "NoVersion"
        SetCommSpeed(12)
        CommFps = 100
    End Sub

    Friend Sub Set_Adc24_MasterPinIndex(ByVal PinsCaps() As String)
        For i As Int32 = 0 To PinsCaps.Length - 1
            If PinsCaps(i) = Adc24_MasterPinName.ToString Then
                Adc24_MasterPinIndex = i
                Adc24_FirstChannelIndex = i + 1
                Exit For
            End If
        Next
    End Sub

    Friend Sub SetPinCaps(ByVal PinsCaps() As String)
        PinsCapsFromIotModule = PinsCaps
        InitDefaultPins()
    End Sub

    Friend Sub InitSlots(ByVal FirstSlot As Int32)
        For Each p As Pin In Pins
            p.Slot = FirstSlot
            FirstSlot += 1
        Next
    End Sub

    ' ==========================================================================================
    '   PUBLIC COMMANDS
    ' ==========================================================================================
    Friend Sub CalibrateZero()
        For Each p As Pin In Pins
            p.CalibrateZero()
        Next
    End Sub

    Friend Function CalibrationNeeded() As Boolean
        For Each p As Pin In Pins
            Select Case p.GetPinType()
                Case Pin.PinTypes.CAP_8, Pin.PinTypes.CAP_16
                    If p.MinVariation > 0 Then
                        Return True
                    End If
            End Select
        Next
        Return False
    End Function

    Friend Sub Calibrate()
        For Each p As Pin In Pins
            p.Calibrate()
        Next
    End Sub

    Friend Sub SetSpeed(ByVal _Speed As Int32)
        StopTimedOperations()
        SyncLock mActionLock
            SetCommSpeed(_Speed)
        End SyncLock
        StartTimedOperations()
    End Sub

    Friend Sub SetOutputPinsToParkingPosition()
        StopTimedOperations()
        SyncLock mActionLock
            For Each p As Pin In Pins
                If p.Direction = Pin.Directions.HostToModule Then
                    ' -------------------------------------------
                    p.Value = p.Value_Parking
                    If Single.IsNaN(p.Value) Then
                        Slots.WriteSlot(p.Slot, p.Value)
                    Else
                        p.ConvertValueToHardwareFormat()
                        Slots.WriteSlot(p.Slot, p.Value_RawUinteger)
                    End If
                End If
            Next
            ' --------------------------------------------------- update phisical module immediately
            DataExchange()
            System.Threading.Thread.Sleep(32)
        End SyncLock
        StartTimedOperations()
    End Sub


    ' ==========================================================================================
    '   PINS
    ' ==========================================================================================
    Friend Pins As List(Of Pin)

    Private Sub InitDefaultPins()
        Pins = New List(Of Pin)
        For i As Int32 = 0 To PinsCapsFromIotModule.Length - 1
            Pins.Add(New Pin(Pin.PinTypes.UNUSED, ModuleId, PinsCapsFromIotModule(i), i))
        Next
    End Sub

    Friend Sub FillTypeCombo(ByVal cmb As ComboBox, ByRef p As Pin)
        cmb.Items.Clear()
        GetSinglePinCaps(p)
        If SinglePinPropsFromIotModule.Prop IsNot Nothing Then
            For i As Int32 = 0 To SinglePinPropsFromIotModule.Prop.Length - 1
                cmb.Items.Add(Pin.PinTypeToString(CType(SinglePinPropsFromIotModule.Prop(i), Theremino_IotHAL.Pin.PinTypes)))
            Next
        Else
            cmb.Items.Add("---")
        End If
    End Sub


    ' SLOW TEST - Valid pin type
    Friend Function TestValidPinType(ByVal PinType As Pin.PinTypes, ByRef p As Pin) As Boolean
        GetSinglePinCaps(p)
        If SinglePinPropsFromIotModule.Prop IsNot Nothing Then
            For i As Int32 = 0 To SinglePinPropsFromIotModule.Prop.Length - 1
                If SinglePinPropsFromIotModule.Prop(i) = PinType Then
                    Return True
                End If
            Next
        Else
            Return False
        End If
    End Function


    ' ##########################################################################################
    ' ******************************************************************************************
    '   COMMUNICATION WITH THE HARDWARE MODULES
    ' ******************************************************************************************
    ' ##########################################################################################

    Private Sub Execution_Error(ByVal ErrorString As String)
        ErrorRate += 0.01F
        Debug.Print("ERROR - " & ErrorString)
    End Sub

    Private Sub Execution_OK()
        ErrorRate *= 0.95F
    End Sub

    ' ==========================================================================================
    '   LOOPS AND COMMUNICATIONS
    ' ==========================================================================================
    Private sw1 As Diagnostics.Stopwatch = New Diagnostics.Stopwatch
    Private rawFps As Single
    Private delayMs As Int32
    Private Sub DataExchange()
        ' ------------------------------------------------------------- if disconnected return
        If Not MyDeviceIsActive Then
            CommFps = 0
            ErrorRate = 100
            Threading.Thread.Sleep(10)
            Return
        End If
        ' ------------------------------------------------------------- ExecuteDataExchange
        ExecuteDataExchange()
        ' ------------------------------------------------------------- CommMillisec
        'CommMillisec = CSng(sw1.Elapsed.TotalMilliseconds)
        CommMillisec = CSng(sw1.ElapsedMilliseconds)
        rawFps = 1000.0F / CommMillisec
        sw1.Reset()
        sw1.Start()
        If CommMillisec >= 1 Then
            SmoothValue(CommFps, rawFps, 1, False, CommFps)
        End If
        ' ------------------------------------------------------------- adaptive FPS
        If mMaxFps < 9999 Then
            delayMs += Math.Sign(rawFps - mMaxFps)
            If delayMs > 100 Then delayMs = 100
            If delayMs <= 0 Then
                delayMs = 0
            Else
                System.Threading.Thread.Sleep(delayMs)
            End If
        End If
    End Sub

    Private Sub ExecuteDataExchange()
        If Not DataExchangeCompleted Then Return
        If PinsPropsProgrammedByIOTHAL Is Nothing Then Return
        ' ----------------------------------------------------------- SLOTS to COM-BUFFER
        Dim i As Int32 = 0
        For Each p As Pin In Pins
            If p.Direction = Pin.Directions.HostToModule Then
                p.ReadValueFromSlot()
                p.ConvertValueToHardwareFormat()
                Select Case p.GetPinType
                    Case Pin.PinTypes.SERVO
                        OutputDataBuffer(i) = p.Value_ServoMicrosec
                    Case Pin.PinTypes.PWM, _
                         Pin.PinTypes.GEN_OUT_FLOAT
                        OutputDataBuffer(i) = p.Value_Normalized
                    Case Else
                        OutputDataBuffer(i) = p.Value_RawUinteger
                End Select
                i += 1
            End If
        Next
        ' ----------------------------------------------------------- Send output pin values to module
        ' ----------------------------------------------------------- every single pin may be paired with a slot
        Dim PinsOut(-1) As Single
        Dim PinCount As Int32 = 0
        For Each p As Pin In Pins
            If p.Direction = Pin.Directions.HostToModule Then
                ReDim Preserve PinsOut(PinCount)
                PinsOut(PinCount) = OutputDataBuffer(PinCount)
                PinCount = PinCount + 1
            End If
        Next

        ' ----------------------------------------------------------- Exchange Request
        DataExchangeCompleted = False
        ' the request is always executed also if there are no output to transfer, 
        ' to force the retrivement of the inputs from the module, if any
        ' XchgModuleReplay event will be fired automatically 
        Dim ModuleProps As String
        ModuleProps = IpAddr
        _NetHALRPC.XchgSingleModuleRequest(ModuleProps, PinsOut)
        '
        ' ----------------------------------------------------------- Wait response or 100 mS timeout
        Static timeout As Stopwatch = New Stopwatch
        Static errorCounter As Int32
        While Not DataExchangeCompleted
            timeout.Start()
            SleepMyThread(1)
            If timeout.ElapsedMilliseconds > 100 Then
                timeout.Reset()
                DataExchangeCompleted = True
                Execution_Error("FastDataExchange timeout")
                errorCounter += 1
                If errorCounter > 20 Then
                    MyDeviceIsActive = False
                End If
                Return
            End If
        End While
        errorCounter = 0
        timeout.Reset()
        Execution_OK()
        ' ----------------------------------------------------------- Test the NON-PROGRAMMED status
        Static NotProgrammedCounter As Int32 = 0
        If ProgrammedStatus = "N" Then
            NotProgrammedCounter += 1
            If NotProgrammedCounter > 50 Then
                MyDeviceIsActive = False
                NotProgrammedCounter = 0
            End If
        Else
            NotProgrammedCounter = 0
        End If
        ' ----------------------------------------------------------- COM-BUFFER to SLOTS
        i = 0
        For Each p As Pin In Pins
            If p.Direction = Pin.Directions.ModuleToHost Then
                If p.GetPinType = Pin.PinTypes.GEN_IN_FLOAT Then
                    p.Value_Normalized_New = InputDataBuffer(i)
                Else
                    If InputDataBuffer(i) < 0 Then InputDataBuffer(i) = 0
                    p.Value_RawUinteger = CUInt(InputDataBuffer(i))
                End If
                i += 1
                p.ConvertValueFromHardwareFormat()
                p.WriteValueToSlot()
            End If
        Next
    End Sub

    Friend Sub SetupModulePins()
        StopTimedOperations()
        SyncLock mActionLock
            ' -------------------------------------------------------- retry 10 times
            For retry As Int32 = 1 To 10
                ' ---------------------------------------------------- send pin types
                Dim ModuleProps As String = IpAddr
                ' ---------------------------------------------------- 
                Dim PinsProps(Pins.Count - 1) As NetHALRPC.PinProps
                ' ---------------------------------------------------- 
                For i As Int32 = 0 To Pins.Count - 1
                    PinsProps(i).Index = i
                    PinsProps(i).Prop = Pins(i).GetPinType
                Next
                PinsPropsProgrammedByIOTHAL = PinsProps
                ' ---------------------------------------------------- ProgModuleRequest
                ProgModuleCompleted = False
                _NetHALRPC.ProgAllModuleRequest(ModuleProps, PinsProps)
                ' ---------------------------------------------------- wait ProgModuleCompleted
                For i As Int32 = 0 To 100
                    SleepMyThread(1)
                    If ProgModuleCompleted Then
                        retry = 9999
                        Exit For
                    End If
                Next
            Next
        End SyncLock
        StartTimedOperations()
        ' ------------------------------------------------------------ send frequency to all the Pwm 
        For i As Int32 = 0 To Pins.Count - 1
            If Pins(i).GetPinType = Pin.PinTypes.PWM Then
                SendPwmPinFrequency(i, Pins(i).PwmFrequency)
            End If
        Next
    End Sub

    Friend Sub WriteNameToHardware()
        StopTimedOperations()
        SyncLock mActionLock
            ' -----------------------------------------------------------------
            ' Retry is in the calling function "Change_SelectedModule_Name" 
            ' -----------------------------------------------------------------
            Dim NameModuleProps As NetHALRPC.NameModulePropsStr
            NameModuleProps.IPAddr = IpAddr
            NameModuleProps.Name = ModuleName
            _NetHALRPC.NameModuleRequest(NameModuleProps)
        End SyncLock
        StartTimedOperations()
    End Sub

    Friend SinglePinPropsFromIotModule As NetHALRPC.GetSinglePinReplayStr = Nothing

    Friend Sub GetSinglePinCaps(ByVal p As Pin)
        StopTimedOperations()
        SyncLock mActionLock
            ' -------------------------------------------------------- retry 10 times
            For retry As Int32 = 1 To 10
                ' ---------------------------------------------------- send pin types
                Dim ModuleProps As String
                ModuleProps = IpAddr
                SinglePinPropsRequestCompleted = False
                _NetHALRPC.GetSinglePinRequest(ModuleProps, CUInt(p.PinIndex))
                ' ---------------------------------------------------- wait ProgModuleCompleted
                For i As Int32 = 0 To 100
                    SleepMyThread(1)
                    If SinglePinPropsRequestCompleted Then
                        retry = 9999
                        Exit For
                    End If
                Next
            Next
        End SyncLock
        StartTimedOperations()
    End Sub


    ' ==================================================================================================
    '   PWM PIN CONFIGURATION 
    ' ==================================================================================================
    Friend Sub SendPwmPinFrequency(ByVal PinId As Int32, ByVal Freq As Single)
        StopTimedOperations()
        SyncLock mActionLock
            ' -------------------------------------------------------- retry 10 times
            For retry As Int32 = 1 To 10
                ' ---------------------------------------------------- ModuleProps
                Dim ModuleProps As String
                ModuleProps = IpAddr
                ' ---------------------------------------------------- PwmPinData
                Dim PwmPinData As NetHALRPC.PwmPinData
                PwmPinData.PinId = CByte(PinId)
                PwmPinData.Frequency = Freq
                ' ---------------------------------------------------- PwmPinDataRequest
                PwmPinDataCompleted = False
                _NetHALRPC.PwmPinDataRequest(ModuleProps, PwmPinData)
                ' ---------------------------------------------------- wait PwmPinDataCompleted
                For i As Int32 = 0 To 100
                    SleepMyThread(1)
                    If PwmPinDataCompleted Then
                        retry = 9999
                        Exit For
                    End If
                Next
            Next
        End SyncLock
        StartTimedOperations()
    End Sub

    ' ==================================================================================================
    '  STEPPER PIN CONFIGURATION  
    ' ==================================================================================================
    Friend Sub SendStepperPinData(ByVal PinId As Int32, _
                                  ByVal StepperSpeedHz As Int32, _
                                  ByVal StepperAccHz_per_mS As Int32)
        StopTimedOperations()
        SyncLock mActionLock
            ' -------------------------------------------------------- retry 10 times
            For retry As Int32 = 1 To 10
                ' ---------------------------------------------------- ModuleProps
                Dim ModuleProps As String
                ModuleProps = IpAddr
                ' ---------------------------------------------------- StepperPinData
                Dim StepperPinData As NetHALRPC.StepperPinData
                StepperPinData.PinId = CByte(PinId)
                StepperPinData.StepperSpeedHz = CUShort(StepperSpeedHz)
                StepperPinData.StepperAccHz_per_mS = CUShort(StepperAccHz_per_mS)
                ' ---------------------------------------------------- StepperPinDataRequest
                StepperPinDataCompleted = False
                _NetHALRPC.StepperPinDataRequest(ModuleProps, StepperPinData)
                ' ---------------------------------------------------- wait StepperPinDataCompleted
                For i As Int32 = 0 To 100
                    SleepMyThread(1)
                    If StepperPinDataCompleted Then
                        retry = 9999
                        Exit For
                    End If
                Next
            Next
        End SyncLock
        StartTimedOperations()
    End Sub

    ' ==================================================================================================
    '  ADC24 CONFIGURATION  
    ' ==================================================================================================
    Friend Sub SendAdc24Configuration(ByVal Adc24ConfigData As NetHALRPC.Adc24ConfigData)
        StopTimedOperations()
        SyncLock mActionLock
            ' -------------------------------------------------------- retry 10 times
            For retry As Int32 = 1 To 10
                ' ---------------------------------------------------- ModuleProps
                Dim ModuleProps As String
                ModuleProps = IpAddr
                ' ---------------------------------------------------- Adc24ConfigRequest
                Adc24ConfigCompleted = False
                _NetHALRPC.Adc24ConfigRequest(ModuleProps, Adc24ConfigData)
                ' ---------------------------------------------------- wait Adc24ConfigCompleted
                For i As Int32 = 0 To 100
                    SleepMyThread(1)
                    If Adc24ConfigCompleted Then
                        retry = 9999
                        Exit For
                    End If
                Next
            Next
        End SyncLock
        StartTimedOperations()
    End Sub



    ' ==================================================================================================
    '   ADC_24 SEND CONFIGURATION and START
    ' ==================================================================================================
    'Private Adc24SpsArray() As Int32 = New Int32() {10, 20, 50, 100, 200, 500, 1000, 2000, _
    '                                                3200, 3840, 4800, 6400, 9600, 19200}

    'Adc24Fsck_KHz = 31,36,42,50,63,83,125,143,167,200,250,333,500,
    '                571,667,800,1000,1333,2000,2286,2667,3200,4000

    Friend Sub Adc24ConfigAndStart()
        Dim Sps As Int32                       ' Samples per second
        Dim MultipleSetups As Boolean
        Dim FiltersSingleCycle As Boolean

        Dim Adc24PreFilterType As Int32        ' Pre filter type 0b000 to 0b111 
        Dim Adc24PostFilterType As Int32       ' Post filter type 0b000 to 0b111 
        Dim Adc24ActiveChannels As Int32       ' The channel is active (bit0 = ch0)
        Dim Adc24BiasedChannels As Int32       ' Channels biased with 1.65 Volt (vcc / 2) 
        Dim Adc24DiffChannels As Int32         ' Differential / SingleEnded channels
        Dim Adc24PseudoChannels As Int32       ' Pseudo channels referred to Input 15
        Dim Adc24ChannelsGains As Int32        ' Gain 1 to 128 (low 3 bits are for channel 0)

        Sps = Pins(Adc24_MasterPinIndex).Adc24Sps


        MultipleSetups = True               ' Every channel's couple uses a different setup
        FiltersSingleCycle = False          ' All filters not using SingleCycle 

        ' ----------- FILTER LIST
        ' Max Speed
        ' Fast
        ' Medium
        ' Slow
        ' Post Max
        ' Post Fast 
        ' Post Medium
        ' Post Slow

        Select Case Pins(Adc24_MasterPinIndex).Adc24Filter
            Case 0 ' Sync3
                Adc24PreFilterType = Convert.ToInt32("010", 2)
                Adc24PostFilterType = Convert.ToInt32("011", 2) ' unused
            Case 1 ' Sync4
                Adc24PreFilterType = Convert.ToInt32("000", 2)
                Adc24PostFilterType = Convert.ToInt32("011", 2) ' unused
            Case 2 ' Sync3 FastSettling
                Adc24PreFilterType = Convert.ToInt32("101", 2)
                Adc24PostFilterType = Convert.ToInt32("011", 2) ' unused
            Case 3 ' Sync4 FastSettling
                Adc24PreFilterType = Convert.ToInt32("100", 2)
                Adc24PostFilterType = Convert.ToInt32("011", 2) ' unused
            Case 4 ' Post1
                Adc24PreFilterType = Convert.ToInt32("111", 2)
                Adc24PostFilterType = Convert.ToInt32("010", 2)
            Case 5 ' Post2
                Adc24PreFilterType = Convert.ToInt32("111", 2)
                Adc24PostFilterType = Convert.ToInt32("011", 2)
            Case 6 ' Post3
                Adc24PreFilterType = Convert.ToInt32("111", 2)
                Adc24PostFilterType = Convert.ToInt32("101", 2)
            Case 7 ' Post4
                Adc24PreFilterType = Convert.ToInt32("111", 2)
                Adc24PostFilterType = Convert.ToInt32("110", 2)
        End Select

        ' ------------------------------------------------------------- FILL MASKS
        Adc24ActiveChannels = 0
        Adc24DiffChannels = 0
        Adc24PseudoChannels = 0
        Adc24ChannelsGains = 0
        Adc24BiasedChannels = 0

        Dim pow2 As Int32 = 1
        Dim pow8 As Int32 = 1
        Dim p As Pin
        For i As Int32 = 0 To 15
            If i + Adc24_FirstChannelIndex >= Pins.Count Then Exit For
            p = Pins(i + Adc24_FirstChannelIndex)
            If p.GetPinType = Pin.PinTypes.ADC_24_CH OrElse _
               p.GetPinType = Pin.PinTypes.ADC_24_CH_B Then
                ' ----------------------------------------------------- If differential couple, set only first pin
                If (p.PinIndex - Adc24_FirstChannelIndex) Mod 2 = 0 Or p.Adc24ChType <> 0 Then
                    Adc24ActiveChannels += pow2
                End If
                ' ----------------------------------------------------- Pseudo
                If p.Adc24ChType = 0 Then Adc24DiffChannels += pow2
                If p.Adc24ChType = 1 Then Adc24PseudoChannels += pow2
                ' ----------------------------------------------------- Gains
                If p.GetPinType <> Pin.PinTypes.UNUSED Then
                    Adc24ChannelsGains = Adc24ChannelsGains Or (p.Adc24ChGain * pow8)
                End If
                ' ----------------------------------------------------- Biased
                If p.Adc24ChBias Then Adc24BiasedChannels += pow2
            End If
            ' ---------------------------------------------------------
            pow2 *= 2
            If (p.PinIndex - Adc24_FirstChannelIndex) Mod 2 = 1 Then
                pow8 *= 8
            End If
        Next

        If Adc24PseudoChannels <> 0 Then
            Adc24BiasedChannels = Adc24BiasedChannels Or Convert.ToInt32("1000000000000000", 2)
        End If

        ' ------------------------------------------------------------- OVERRIDE MASKS
        '' active
        'Adc24ActiveChannels = Convert.ToInt32("0000000000000001", 2)
        '' Combo - Diff Pseudo Single
        'Adc24DiffChannels = Convert.ToInt32("0000000000000001", 2)
        'Adc24PseudoChannels = Convert.ToInt32("0000000000000000", 2)
        '' Combo - Gain 1 2 4 8 16 32 64 128
        'Adc24ChannelsGains = Convert.ToInt32("111111111111111111111111", 2) ' 000=1 001=2 010=4 011=8 100=16 ... 110=64 111=128
        '' Check - Biased
        'Adc24BiasedChannels = Convert.ToInt32("0000000000000000", 2)

        ' --------------------------------------------------------------------------------
        '   ADC_24 CONFIG 
        ' --------------------------------------------------------------------------------
        Dim Adc24ConfigData As NetHALRPC.Adc24ConfigData
        Adc24ConfigData.Sps = CUShort(Sps)
        Adc24ConfigData.MultipleSetups = MultipleSetups
        Adc24ConfigData.FiltersSingleCycle = FiltersSingleCycle
        Adc24ConfigData.Adc24PreFilterType = CByte(Adc24PreFilterType)
        Adc24ConfigData.Adc24PostFilterType = CByte(Adc24PostFilterType)
        Adc24ConfigData.Adc24ActiveChannels = CUShort(Adc24ActiveChannels)
        Adc24ConfigData.Adc24BiasedChannels = CUShort(Adc24BiasedChannels)
        Adc24ConfigData.Adc24DiffChannels = CUShort(Adc24DiffChannels)
        Adc24ConfigData.Adc24PseudoChannels = CUShort(Adc24PseudoChannels)
        Dim b(2) As Byte
        b(2) = CByte((Adc24ChannelsGains >> 16) And 255)
        b(1) = CByte((Adc24ChannelsGains >> 8) And 255)
        b(0) = CByte(Adc24ChannelsGains And 255)
        Adc24ConfigData.Adc24ChannelsGains = b
        SendAdc24Configuration(Adc24ConfigData)
    End Sub

End Class

