﻿
' ==================================================================================================
'   CLASS PIN
' ==================================================================================================
Class Pin

    Friend Enum Directions As Int32
        Unused = 0
        HostToModule = 1
        ModuleToHost = 2
    End Enum

    Friend Enum PinTypes As Byte
        UNUSED = 0
        ' -------------------- OUT
        DIG_OUT = 1
        'PWM_8 = 2
        'PWM_16 = 3
        'SERVO_8 = 4
        'SERVO_16 = 5
        STEPPER = 6
        'PWM_FAST = 7
        DAC_8 = 8
        SERVO = 10
        'SERVO_100Hz = 11
        PWM = 15

        GEN_OUT_8 = 90
        GEN_OUT_16 = 91
        GEN_OUT_24 = 92
        GEN_OUT_FLOAT = 93
        ' -------------------- IN
        DIG_IN = 129
        DIG_IN_PU = 130

        ADC_8 = 131
        ADC_16 = 132

        ADC_24 = 173
        'ADC_24_DIN = 174
        'ADC_24_DOUT = 175
        ADC_24_CH = 176
        ADC_24_CH_B = 177

        CAP_8 = 133
        CAP_16 = 134

        COUNTER = 140
        COUNTER_PU = 141
        PERIOD = 144
        PERIOD_PU = 145

        STEPPER_DIR = 165 ' STEPPER_DIR is an hardware output but it sends steps-from-destination to USB

        ENCODER_A = 180
        ENCODER_A_PU = 181
        ENCODER_B = 182
        ENCODER_B_PU = 183

        GEN_IN_8 = 190
        GEN_IN_16 = 191
        GEN_IN_24 = 192
        GEN_IN_FLOAT = 193
    End Enum


    Friend Shared Function StringToPinType(ByVal PinTypeString As String) As PinTypes
        Select Case UCase(PinTypeString.Trim)
            Case "UNUSED" : Return PinTypes.UNUSED
                ' ------------------------------------------------- OUT
            Case "DIG_OUT" : Return PinTypes.DIG_OUT
            Case "DAC_8" : Return PinTypes.DAC_8
            Case "SERVO" : Return PinTypes.SERVO
            Case "PWM" : Return PinTypes.PWM
            Case "STEPPER" : Return PinTypes.STEPPER
            Case "GEN_OUT_8" : Return PinTypes.GEN_OUT_8
            Case "GEN_OUT_16" : Return PinTypes.GEN_OUT_16
            Case "GEN_OUT_24" : Return PinTypes.GEN_OUT_24
            Case "GEN_OUT_FLOAT" : Return PinTypes.GEN_OUT_FLOAT

                ' ------------------------------------------------- IN
            Case "DIG_IN" : Return PinTypes.DIG_IN
            Case "DIG_IN_PU" : Return PinTypes.DIG_IN_PU

            Case "CAP_8" : Return PinTypes.CAP_8
            Case "CAP_16" : Return PinTypes.CAP_16

            Case "ADC_8" : Return PinTypes.ADC_8
            Case "ADC_16" : Return PinTypes.ADC_16
            Case "ADC_24" : Return PinTypes.ADC_24
            Case "ADC_24_CH" : Return PinTypes.ADC_24_CH
            Case "ADC_24_CH_B" : Return PinTypes.ADC_24_CH_B

            Case "COUNTER" : Return PinTypes.COUNTER
            Case "COUNTER_PU" : Return PinTypes.COUNTER_PU
            Case "PERIOD" : Return PinTypes.PERIOD
            Case "PERIOD_PU" : Return PinTypes.PERIOD_PU

            Case "STEPPER_DIR" : Return PinTypes.STEPPER_DIR

            Case "ENCODER_A" : Return PinTypes.ENCODER_A
            Case "ENCODER_A_PU" : Return PinTypes.ENCODER_A_PU
            Case "ENCODER_B" : Return PinTypes.ENCODER_B
            Case "ENCODER_B_PU" : Return PinTypes.ENCODER_B_PU

            Case "GEN_IN_8" : Return PinTypes.GEN_IN_8
            Case "GEN_IN_16" : Return PinTypes.GEN_IN_16
            Case "GEN_IN_24" : Return PinTypes.GEN_IN_24
            Case "GEN_IN_FLOAT" : Return PinTypes.GEN_IN_FLOAT

        End Select
        Return Nothing
    End Function

    Friend Shared Function PinTypeToString(ByVal pintype As PinTypes) As String
        Dim s As String = ""
        Select Case pintype
            Case PinTypes.UNUSED : s = "UNUSED"
                ' ------------------------------------------------- OUT
            Case PinTypes.DIG_OUT : s = "DIG_OUT"
            Case PinTypes.DAC_8 : s = "DAC_8"
            Case PinTypes.SERVO : s = "SERVO"
            Case PinTypes.PWM : s = "PWM"
            Case PinTypes.STEPPER : s = "STEPPER"
            Case PinTypes.GEN_OUT_8 : s = "GEN_OUT_8"
            Case PinTypes.GEN_OUT_16 : s = "GEN_OUT_16"
            Case PinTypes.GEN_OUT_24 : s = "GEN_OUT_24"
            Case PinTypes.GEN_OUT_FLOAT : s = "GEN_OUT_FLOAT"

                ' ------------------------------------------------- IN
            Case PinTypes.DIG_IN : s = "DIG_IN"
            Case PinTypes.DIG_IN_PU : s = "DIG_IN_PU"

            Case PinTypes.CAP_8 : s = "CAP_8"
            Case PinTypes.CAP_16 : s = "CAP_16"

            Case PinTypes.ADC_8 : s = "ADC_8"
            Case PinTypes.ADC_16 : s = "ADC_16"
            Case PinTypes.ADC_24 : s = "ADC_24"
            Case PinTypes.ADC_24_CH : s = "ADC_24_CH"
            Case PinTypes.ADC_24_CH_B : s = "ADC_24_CH_B"

            Case PinTypes.COUNTER : s = "COUNTER"
            Case PinTypes.COUNTER_PU : s = "COUNTER_PU"
            Case PinTypes.PERIOD : s = "PERIOD"
            Case PinTypes.PERIOD_PU : s = "PERIOD_PU"

            Case PinTypes.STEPPER_DIR : s = "STEPPER_DIR"

            Case PinTypes.ENCODER_A : s = "ENCODER_A"
            Case PinTypes.ENCODER_A_PU : s = "ENCODER_A_PU"
            Case PinTypes.ENCODER_B : s = "ENCODER_B"
            Case PinTypes.ENCODER_B_PU : s = "ENCODER_B_PU"

            Case PinTypes.GEN_IN_8 : s = "GEN_IN_8"
            Case PinTypes.GEN_IN_16 : s = "GEN_IN_16"
            Case PinTypes.GEN_IN_24 : s = "GEN_IN_24"
            Case PinTypes.GEN_IN_FLOAT : s = "GEN_IN_FLOAT"

        End Select

        Return CamelCaseString(s)
    End Function



    Friend ModuleId As Int32
    Friend PinName As String
    Friend PinIndex As Int32
    Private PinType As PinTypes
    Friend Direction As Directions
    ' ------------------------------------------- slot 0 to 999 starting from the pin near to module
    Friend Slot As Int32
    ' ------------------------------------------- byte index in the buffer ( HostToModule or ModuleToHost )
    'Friend CommByteIndex As Int32

    ' ------------------------------------------- all pins
    Friend Value_RawInteger As Int32
    Friend Value_RawUinteger As UInt32
    Friend Value_RawUinteger_Old As UInt32
    Friend Value_RawUinteger_Zero As UInt32
    Friend Value_RawSmoothed As Single
    Friend Value_Uinteger_Max As UInt32
    Friend Value_Uinteger_Min As UInt32
    Friend Value As Single
    Friend Value_Max As Single
    Friend Value_Min As Single
    Friend Value_Parking As Single
    Friend Value_Normalized As Single = 0
    Friend Value_Normalized_New As Single
    Friend Value_ServoMicrosec As Single
    Friend Value_Zero As Single
    Friend AdaptiveSpeed As Boolean
    Friend ResponseSpeed As Int32
    Private OldTimeMilliseconds As Double
    Private NewTimeMilliseconds As Double
    ' ------------------------------------------ Servo and Pwm specific
    Friend MaxTime As Single
    Friend MinTime As Single
    Friend LogResponse As Boolean
    ' ------------------------------------------ Stepper specific
    Friend MaxSpeed As Single
    Friend MaxAcc As Single
    Friend StepsPerMillimeter As Single
    Friend StepperSpeed_Hz As Int32
    Friend StepperAcc_Hz_per_mS As Int32
    ' ------------------------------------------ Pwm specific
    Friend PwmFrequency As Single
    ' ------------------------------------------ Counter and Period specific
    Friend ConvertToFreq As Boolean
    Friend MaxFreq As Single
    Friend MinFreq As Single
    ' ------------------------------------------ Touch specific
    Friend MinVariation As Single
    Friend ProportionalArea As Single
    Friend MeanValue As Single
    Private oldDelta1 As Single
    Private oldDelta2 As Single
    Private oldDelta3 As Single
    Private Velocity As Single
    '
    ' ------------------------------------------ Adc_24 specific
    Friend Adc24Sps As Int32                  ' Samples per second
    Friend Adc24Filter As Int32               ' Filters 0 to 7
    '
    ' ------------------------------------------ Adc_24_channel specific
    Friend Adc24ChType As Int32               ' (0 to 2) Diff / Pseudo / Single ended
    Friend Adc24ChGain As Int32               ' (0 to 7) 1,2,4,8,16,32,64,128
    Friend Adc24ChBias As Boolean             ' Bias on off  
    '
    '
    Friend Freq As Single
    '
    Friend Sub New(ByVal _pintype As PinTypes, _
                   ByVal _moduleid As Int32, _
                   ByVal _pinname As String, _
                   ByVal _pinindex As Int32)

        ModuleId = _moduleid
        PinName = _pinname
        PinIndex = _pinindex
        SetPinType(_pintype)
    End Sub

    Friend Function GetPinType() As PinTypes
        Return PinType
    End Function


    ' ==================================================================================================
    '   DEFAULTS
    ' ==================================================================================================
    Friend Sub SetPinType(ByVal _pintype As PinTypes)
        If PinType = _pintype Then Return
        PinType = _pintype
        '
        Value_Parking = 0
        Value_Max = 1000
        Value_Min = 0
        MaxFreq = 1000
        MinFreq = 0
        AdaptiveSpeed = False
        ResponseSpeed = 100
        LogResponse = False
        ' --------------------------------------------------------- defaults
        Select Case PinType

            ' ----------------------------------------------------- outputs
            Case PinTypes.DIG_OUT
                Direction = Directions.HostToModule

            Case PinTypes.SERVO
                Value_Parking = NAN_Sleep
                MaxTime = 2500
                MinTime = 500
                Direction = Directions.HostToModule

            Case PinTypes.DAC_8
                Direction = Directions.HostToModule

            Case PinTypes.STEPPER
                MaxSpeed = 200
                MaxAcc = 50
                StepsPerMillimeter = 200
                Direction = Directions.HostToModule

            Case PinTypes.PWM
                PwmFrequency = 100
                Direction = Directions.HostToModule

            Case PinTypes.GEN_OUT_8, PinTypes.GEN_OUT_16, _
                 PinTypes.GEN_OUT_24, PinTypes.GEN_OUT_FLOAT
                Direction = Directions.HostToModule

                ' ------------------------------------------------- inputs
            Case PinTypes.DIG_IN, PinTypes.DIG_IN_PU
                Value_Parking = 0
                Direction = Directions.ModuleToHost

            Case PinTypes.CAP_8, PinTypes.CAP_16
                MinVariation = 20
                ProportionalArea = 350
                MeanValue = 0
                ResponseSpeed = 100
                Direction = Directions.ModuleToHost

            Case PinTypes.ADC_8
                ResponseSpeed = 30
                Direction = Directions.ModuleToHost

            Case PinTypes.ADC_16
                ResponseSpeed = 30
                Direction = Directions.ModuleToHost

            Case PinTypes.ADC_24
                Adc24Sps = 100
                Adc24Filter = 0
                Direction = Directions.ModuleToHost

            Case PinTypes.ADC_24_CH
                Adc24ChType = 0
                Adc24ChGain = 0
                Adc24ChBias = False
                Direction = Directions.ModuleToHost

            Case PinTypes.ADC_24_CH_B
                Adc24ChType = 0
                Adc24ChGain = 0
                Adc24ChBias = False
                Direction = Directions.Unused

            Case PinTypes.COUNTER, PinTypes.COUNTER_PU
                Direction = Directions.ModuleToHost

            Case PinTypes.PERIOD, PinTypes.PERIOD_PU
                ResponseSpeed = 30
                Direction = Directions.ModuleToHost

            Case PinTypes.GEN_IN_8, PinTypes.GEN_IN_16, _
                 PinTypes.GEN_IN_24, PinTypes.GEN_IN_FLOAT
                Direction = Directions.ModuleToHost

            Case PinTypes.STEPPER_DIR
                Direction = Directions.ModuleToHost

            Case PinTypes.ENCODER_A, PinTypes.ENCODER_A_PU
                Direction = Directions.ModuleToHost
            Case PinTypes.ENCODER_B, PinTypes.ENCODER_B_PU
                Direction = Directions.Unused

            Case PinTypes.UNUSED
                MaxTime = 1
                MinTime = 0
                Value_Parking = 0
                Direction = Directions.Unused

            Case Else
                PinType = PinTypes.UNUSED
                MaxTime = 1
                MinTime = 0
                Value_Parking = 0
                Direction = Directions.Unused
        End Select

    End Sub


    ' ==================================================================================================
    '  OUTPUTS  ( from SLOT to HARDWARE )
    ' ==================================================================================================
    Friend Sub ReadValueFromSlot()
        Value = Slots.ReadSlot(Slot)
    End Sub

    Friend Sub ConvertValueToHardwareFormat()
        Select Case PinType
            Case PinTypes.DIG_OUT, PinTypes.PWM, PinTypes.SERVO, PinTypes.DAC_8, _
                PinTypes.GEN_OUT_8, PinTypes.GEN_OUT_16, PinTypes.GEN_OUT_24
                ' -------------------------------------------------------------- min and max
                If Value_Max <> Value_Min Then
                    Value_Normalized_New = (Value - Value_Min) / (Value_Max - Value_Min)
                Else
                    Value_Normalized_New = Value
                End If
                ' -------------------------------------------------------------- limit to 0..1
                If Value_Normalized_New < 0 Then Value_Normalized_New = 0
                If Value_Normalized_New > 1 Then Value_Normalized_New = 1
                ' -------------------------------------------------------------- exponential passing by 0 and 1
                '                                                                a little less that the log base ( 2.7182 ) 
                If LogResponse AndAlso PinType = PinTypes.PWM Then
                    Value_Normalized_New = CSng(Value_Normalized_New ^ 2)
                End If
                ' -------------------------------------------------------------- speed 
                SmoothValue(Value_Normalized, _
                            Value_Normalized_New, _
                            ResponseSpeed, _
                            AdaptiveSpeed, _
                            InOutModules(ModuleId).CommFps)

            Case PinTypes.GEN_OUT_FLOAT
                ' -------------------------------------------------------------- min and max
                If Value_Max <> Value_Min Then
                    Value_Normalized_New = (Value - Value_Min) / (Value_Max - Value_Min)
                Else
                    Value_Normalized_New = Value
                End If
                ' -------------------------------------------------------------- speed 
                SmoothValue(Value_Normalized, _
                            Value_Normalized_New, _
                            ResponseSpeed, _
                            AdaptiveSpeed, _
                            InOutModules(ModuleId).CommFps)
        End Select

        Select Case PinType
            Case PinTypes.DIG_OUT
                If Single.IsNaN(Value_Normalized) Then
                    Value_Normalized = 0
                End If
                If Value_Normalized > 0.5 Then
                    Value_RawUinteger = 1
                Else
                    Value_RawUinteger = 0
                End If

            Case PinTypes.SERVO
                If Single.IsNaN(Value_Normalized) Then
                    Value_Normalized = If(Value_Max > Value_Min, 0, 1)
                    Value_ServoMicrosec = Value_Normalized * 4000
                Else
                    Value_ServoMicrosec = (Value_Normalized * (MaxTime - MinTime) + MinTime)
                End If

            Case PinTypes.PWM
                If Single.IsNaN(Value_Normalized) Then Value_Normalized = 0

            Case PinTypes.DAC_8
                If Single.IsNaN(Value_Normalized) Then
                    Value_Normalized = 0
                    Value_RawUinteger = 0
                Else
                    Value_RawUinteger = CUInt(255.0F * Value_Normalized)
                End If
                If Value_RawUinteger < 0 Then Value_RawUinteger = 0
                If Value_RawUinteger > 255 Then Value_RawUinteger = 255

            Case PinTypes.GEN_OUT_8
                If Single.IsNaN(Value_Normalized) Then
                    Value_Normalized = If(Value_Max > Value_Min, 0, 1)
                    Value_RawUinteger = 0
                Else
                    Value_RawUinteger = CUInt(255.0F * Value_Normalized)
                End If
                If Value_RawUinteger < 0 Then Value_RawUinteger = 0
                If Value_RawUinteger > 255 Then Value_RawUinteger = 255

            Case PinTypes.GEN_OUT_16
                If Single.IsNaN(Value_Normalized) Then
                    Value_Normalized = If(Value_Max > Value_Min, 0, 1)
                    Value_RawUinteger = 0
                Else
                    Value_RawUinteger = CUInt(65535.0F * Value_Normalized)
                End If
                If Value_RawUinteger < 0 Then Value_RawUinteger = 0
                If Value_RawUinteger > 65535 Then Value_RawUinteger = 65535

            Case PinTypes.GEN_OUT_24
                If Single.IsNaN(Value_Normalized) Then
                    Value_Normalized = If(Value_Max > Value_Min, 0, 1)
                    Value_RawUinteger = 0
                Else
                    Value_RawUinteger = CUInt(16777215.0F * Value_Normalized)
                End If
                If Value_RawUinteger < 0 Then Value_RawUinteger = 0
                If Value_RawUinteger > 16777215 Then Value_RawUinteger = 16777215

            Case PinTypes.GEN_OUT_FLOAT
                If Single.IsNaN(Value_Normalized) Then
                    Value_Normalized = If(Value_Max > Value_Min, 0, 1)
                End If
                Value_Normalized *= 1000

        End Select
    End Sub


    ' ==================================================================================================
    '  INPUTS  ( from HARDWARE to SLOT )
    ' ==================================================================================================
    Friend Sub ConvertValueFromHardwareFormat()

        Select Case PinType
            Case PinTypes.DIG_IN, PinTypes.DIG_IN_PU
                Value_RawUinteger = If(Value_RawUinteger <> 0, 1UI, 0UI)
                Value_Normalized_New = Value_RawUinteger

            Case PinTypes.CAP_8

            Case PinTypes.CAP_16


            Case PinTypes.ADC_8
                Value_Normalized_New = Value_RawUinteger / 255.0F

            Case PinTypes.ADC_16
                Value_Normalized_New = Value_RawUinteger / 65535.0F

                'Case COUNTER COUNTER_PU ' <-- Nothing to do
                'Case PERIOD PERIOD_PU ' <-- Nothing to do

            Case PinTypes.ADC_24
                Value_Normalized_New = Value_RawUinteger / 255.0F

                'Case PinTypes.ADC_24_DOUT
                '    Value_Normalized_New = Value_RawUinteger / 255.0F


            Case PinTypes.ADC_24_CH
                Value_Normalized_New = Value_RawUinteger / 16777215.0F
            Case PinTypes.ADC_24_CH_B

                'Case PinTypes.ENCODER_A, PinTypes.ENCODER_A_PU
                'Case PinTypes.ENCODER_B, PinTypes.ENCODER_B_PU ' <-- Nothing to do

            Case PinTypes.GEN_IN_8
                Value_Normalized_New = Value_RawUinteger / 255.0F
            Case PinTypes.GEN_IN_16
                Value_Normalized_New = Value_RawUinteger / 65535.0F
            Case PinTypes.GEN_IN_24
                Value_Normalized_New = Value_RawUinteger / 16777215.0F
            Case PinTypes.GEN_IN_FLOAT
                Value_Normalized_New = Value_Normalized_New / 1000.0F

        End Select


        Select Case PinType

            Case PinTypes.DIG_IN, PinTypes.DIG_IN_PU, PinTypes.ADC_8, PinTypes.ADC_16
                ' ------------------------------------------------------------------ limit to 0..1
                If Value_Normalized_New < 0 Then Value_Normalized_New = 0
                If Value_Normalized_New > 1 Then Value_Normalized_New = 1
                ' -------------------------------------------------------------- speed 
                SmoothValue(Value_Normalized, _
                            Value_Normalized_New, _
                            ResponseSpeed, _
                            AdaptiveSpeed, _
                            InOutModules(ModuleId).CommFps)
                ' ------------------------------------------------------------------ min and max
                Value = Value_Normalized * (Value_Max - Value_Min) + Value_Min


            Case PinTypes.ADC_24
                Value = Value_Normalized_New * 1000.0F
                Value_Normalized = Value_Normalized_New

            Case PinTypes.ADC_24_CH
                ' -------------------------------------------------------------- speed 
                SmoothValue(Value_Normalized, _
                            Value_Normalized_New, _
                            ResponseSpeed, _
                            AdaptiveSpeed, _
                            InOutModules(ModuleId).CommFps)
                ' -------------------------------------------------------------- min and max
                Value = Value_Normalized * (Value_Max - Value_Min) + Value_Min
            Case PinTypes.ADC_24_CH_B
                '

            Case PinTypes.CAP_8, PinTypes.CAP_16
                ' -------------------------------------------------------------- speed before trigger
                If PinType = PinTypes.CAP_8 Then
                    SmoothValue(Value_RawSmoothed, _
                                Value_RawUinteger * (1000.0F / 255.0F), _
                                ResponseSpeed, _
                                AdaptiveSpeed, _
                                InOutModules(ModuleId).CommFps)
                Else
                    SmoothValue(Value_RawSmoothed, _
                                Value_RawUinteger * (1000.0F / 65535.0F), _
                                ResponseSpeed, _
                                AdaptiveSpeed, _
                                InOutModules(ModuleId).CommFps)
                End If

                If MinVariation > 0 Then
                    ' -------------------------------------------
                    '   CAP TOUCH KEYS
                    ' -------------------------------------------

                    ' ------------------------------------------------------------------ calculate delta
                    Dim delta As Single = MeanValue - Value_RawSmoothed

                    ' ------------------------------------------------------------------ increase min value
                    If delta < 0 Then
                        MeanValue += (Value_RawSmoothed - MeanValue) * 0.01F
                        delta = 0
                    End If
                    ' ------------------------------------------------------------------ MinVariation reduced for sliders
                    If ProportionalArea > 0 Then
                        delta -= MinVariation / 100
                    Else
                        delta -= MinVariation
                    End If
                    ' ------------------------------------------------------------------
                    If ProportionalArea > 0 Then
                        ' -------------------------------------------------------------- proportional limited to 0..1
                        Value_Normalized = delta / ProportionalArea
                        If Value_Normalized < 0 Then Value_Normalized = 0
                        If Value_Normalized > 1 Then Value_Normalized = 1
                    ElseIf ProportionalArea = 0 Then
                        ' -------------------------------------------------------------- on/off (0 or 1)
                        Value_Normalized = If(delta > 0, 1, 0)
                    Else
                        ' -------------------------------------------------------------- with velocity
                        If delta > 0 Then
                            If Velocity = 0 Then
                                Velocity = (delta - oldDelta3) / -ProportionalArea
                                If Velocity > 1 Then Velocity = 1
                                Velocity = Velocity * Velocity
                            End If
                        Else
                            Velocity = 0
                        End If
                        oldDelta3 = oldDelta2
                        oldDelta2 = oldDelta1
                        oldDelta1 = delta
                        Value_Normalized = Velocity
                    End If

                Else
                    ' -------------------------------------------
                    '   RAW CAPACITIVE SENSORS
                    ' -------------------------------------------
                    Value_Normalized = (-MinVariation - Value_RawSmoothed) / ProportionalArea
                    If Value_Normalized < 0 Then Value_Normalized = 0
                    If Value_Normalized > 1 Then Value_Normalized = 1

                End If
                ' ---------------------------------------------------------------------- min and max
                Value = Value_Normalized * (Value_Max - Value_Min) + Value_Min
                ' ---------------------------------------------------------------------- values min and max for calibrate Inhibit
                If Value_RawSmoothed > Value_Uinteger_Max Then Value_Uinteger_Max = CUInt(Value_RawSmoothed * 10)
                If Value_RawSmoothed < Value_Uinteger_Min Then Value_Uinteger_Min = CUInt(Value_RawSmoothed * 10)


            Case PinTypes.COUNTER, PinTypes.COUNTER_PU
                If ConvertToFreq Then
                    ' ------------------------------------------------------------------ delta time
                    NewTimeMilliseconds = FreeRunningTimer.Elapsed.TotalMilliseconds
                    Dim dt As Double = NewTimeMilliseconds - OldTimeMilliseconds
                    ' ------------------------------------------------------------------ update every 100 mS
                    Dim SamplingTimeMs As Double = 100000 / MaxFreq
                    If SamplingTimeMs > 1000 Then SamplingTimeMs = 1000
                    If SamplingTimeMs < 10 Then SamplingTimeMs = 10
                    If dt >= SamplingTimeMs Then
                        ' -------------------------------------------------------------- delta counts
                        Dim diff As UInt32
                        If Value_RawUinteger >= Value_RawUinteger_Old Then
                            diff = Value_RawUinteger - Value_RawUinteger_Old
                        Else
                            diff = (&HFFFFUI - Value_RawUinteger_Old) + Value_RawUinteger + 1UI
                        End If
                        Value_RawUinteger_Old = Value_RawUinteger
                        ' -------------------------------------------------------------- frequency
                        If Value_RawUinteger > 0 AndAlso dt > 0 Then
                            Freq = CSng(1000 * diff / dt)
                        Else
                            Freq = 0
                        End If
                        OldTimeMilliseconds = NewTimeMilliseconds
                    End If
                    ' ------------------------------------------------------------------ range
                    Dim range As Double = MaxFreq - MinFreq
                    If range < 0.1 Then range = 0.1
                    Value_Normalized_New = CSng((Freq - MinFreq) / range)
                    ' ------------------------------------------------------------------ speed
                    SmoothValue(Value_Normalized, _
                                Value_Normalized_New, _
                                ResponseSpeed, _
                                AdaptiveSpeed, _
                                InOutModules(ModuleId).CommFps)
                    ' ------------------------------------------------------------------ min and max
                    Value = Value_Normalized * (Value_Max - Value_Min) + Value_Min
                Else
                    Value_Normalized_New = Value_RawUinteger / 65535.0F
                    Value_Normalized = Value_Normalized_New
                    Value = Value_RawUinteger
                End If

            Case PinTypes.PERIOD, PinTypes.PERIOD_PU
                If ConvertToFreq Then
                    ' -------------------------------------------------------------- range
                    If Value_RawUinteger > 0 Then
                        ' ---------------------------------------------------------- translate to Hertz
                        'Freq = 1.0E+7F / Value_RawUinteger    ' timerRead()   
                        'Freq = 1000000.0F / Value_RawUinteger ' Micros()
                        Freq = 2.4E+8F / Value_RawUinteger     ' XTHAL_GET_CCOUNT()
                    Else
                        Freq = 0
                    End If
                    Dim range As Double = MaxFreq - MinFreq
                    If range < 0.1 Then range = 0.1
                    ' -------------------------------------------------------------- if value is more than 20 KHz then use old value
                    If Freq <= 20000 Then
                        Value_Normalized_New = CSng((Freq - MinFreq) / range)
                    End If
                    ' -------------------------------------------------------------- speed 
                    SmoothValue(Value_Normalized, _
                                Value_Normalized_New, _
                                ResponseSpeed, _
                                AdaptiveSpeed, _
                                InOutModules(ModuleId).CommFps)
                    ' -------------------------------------------------------------- min and max
                    Value = Value_Normalized * (Value_Max - Value_Min) + Value_Min
                Else
                    ' -------------------------------------------------------------- translate to microseconds
                    Value_Normalized_New = Value_RawUinteger * 1000000 / 2.4E+8F
                    ' -------------------------------------------------------------- if value is less than 50 uS (>20KHz) then use old value
                    If Value_Normalized_New >= 50 Then
                        Value_Normalized = Value_Normalized_New
                    End If
                    Value = Value_Normalized
                End If

            Case PinTypes.ENCODER_A, PinTypes.ENCODER_A_PU
                Value_Normalized_New = Value_RawUinteger / 65535.0F
                Value_Normalized = Value_Normalized_New
                Value = Value_RawUinteger

            Case PinTypes.GEN_IN_8, PinTypes.GEN_IN_16, _
                 PinTypes.GEN_IN_24, PinTypes.GEN_IN_FLOAT
                ' -------------------------------------------------------------- speed 
                SmoothValue(Value_Normalized, _
                            Value_Normalized_New, _
                            ResponseSpeed, _
                            AdaptiveSpeed, _
                            InOutModules(ModuleId).CommFps)
                ' -------------------------------------------------------------- min and max
                Value = Value_Normalized * (Value_Max - Value_Min) + Value_Min

        End Select
    End Sub

    Friend Sub WriteValueToSlot()
        Slots.WriteSlot(Slot, Value)
    End Sub


    ' ==================================================================================================
    '   STEPPER PIN CONFIGURATION 
    ' ==================================================================================================
    Friend Sub SendStepperParamsToHardware()
        StepperSpeed_Hz = CInt(MaxSpeed * StepsPerMillimeter / 60)
        StepperAcc_Hz_per_mS = CInt(MaxAcc * StepsPerMillimeter / 1000)
        If StepperSpeed_Hz > 65535 Then StepperSpeed_Hz = 65535
        If StepperAcc_Hz_per_mS > 65535 Then StepperAcc_Hz_per_mS = 65535
        InOutModules(ModuleId).SendStepperPinData(PinIndex, StepperSpeed_Hz, StepperAcc_Hz_per_mS)
    End Sub


    ' ==================================================================================================
    '   HELPER FUNCTIONS
    ' ==================================================================================================
    Friend Sub Calibrate()
        Select Case PinType
            Case PinTypes.CAP_8, PinTypes.CAP_16
                ' ----------------------------------------------- calibrate cap input
                MeanValue = Value_RawSmoothed
        End Select
    End Sub

    Friend Sub CalibrateZero()
        Value_RawUinteger_Zero = Value_RawUinteger
        Value_Zero = Value
    End Sub

    Friend Function GetValueString() As String
        Dim ci As Globalization.CultureInfo = New Globalization.CultureInfo("en-GB")
        Select Case PinType

            ' ------------------------------------------------------------------------- OUTPUTS
            Case PinTypes.DIG_OUT
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        " Slot: " & Value.ToString("0") & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                        "  Raw: " & Value_RawUinteger.ToString("0")

            Case PinTypes.SERVO
                If Single.IsNaN(Value) Then
                    Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        " Slot: " & "Sleep" & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci)
                Else
                    Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        " Slot: " & Value.ToString("0") & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                        " uSec: " & Value_ServoMicrosec.ToString("0.00", ci)
                End If

            Case PinTypes.DAC_8
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        " Slot: " & Value.ToString("0") & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                        "  Raw: " & Value_RawUinteger.ToString("0")

            Case PinTypes.STEPPER
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        " Slot: " & Value.ToString("0") & vbCrLf & _
                        " Filt: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                        " Step: " & (Convert.ToInt64(Value_RawUinteger) - &H80000000UI).ToString("0") & vbCrLf & _
                        "   Hz: " & StepperSpeed_Hz.ToString("0") & vbCrLf & _
                        "Hz/mS: " & StepperAcc_Hz_per_mS.ToString("0")

            Case PinTypes.PWM
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        " Slot: " & Value.ToString("0") & vbCrLf & _
                        " Filt: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                        "   Hz: " & PwmFrequency.ToString("0")

            Case PinTypes.GEN_OUT_8, PinTypes.GEN_OUT_16, PinTypes.GEN_OUT_24
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        " Slot: " & Value.ToString("0") & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                        "  Raw: " & Value_RawUinteger.ToString("0")

            Case PinTypes.GEN_OUT_FLOAT
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        " Slot: " & Value.ToString("0.00", ci) & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf

                ' --------------------------------------------------------------------- INPUTS 
            Case PinTypes.DIG_IN, PinTypes.DIG_IN_PU
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        "  Raw: " & Value_RawUinteger.ToString("0") & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                        " Slot: " & Value.ToString("0.00", ci)

            Case PinTypes.CAP_8, PinTypes.CAP_16
                If MinVariation > 0 Then
                    Return PinTypeToString(PinType) & vbCrLf & _
                                    "--------------" & vbCrLf & _
                                    "   Raw: " & Value_RawUinteger.ToString("000") & vbCrLf & _
                                    " Smoot: " & Value_RawSmoothed.ToString("0.00") & vbCrLf & _
                                    "  Mean: " & MeanValue.ToString("0.00") & vbCrLf & _
                                    "  Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                                    "  Slot: " & Value.ToString("0.00", ci)
                Else
                    Return PinTypeToString(PinType) & vbCrLf & _
                                    "--------------" & vbCrLf & _
                                    "   Raw: " & Value_RawUinteger.ToString("000") & vbCrLf & _
                                    " Smoot: " & Value_RawSmoothed.ToString("0.00") & vbCrLf & _
                                    "  Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                                    "  Slot: " & Value.ToString("0.00", ci)
                End If

            Case PinTypes.ADC_8, PinTypes.ADC_16
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        "  Raw: " & Value_RawUinteger.ToString("000") & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                        " Slot: " & Value.ToString("0.00", ci)

            Case PinTypes.ADC_24
                Return PinTypeToString(PinType) & vbCrLf & _
                       "--------------" & vbCrLf & _
                       "  Raw: " & Value_RawUinteger.ToString("000") & vbCrLf & _
                       " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                       " Slot: " & Value.ToString("0.00", ci)

            Case PinTypes.ADC_24_CH
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        "  Raw: " & Value_RawUinteger.ToString("00000000") & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                        " Slot: " & Value.ToString("0.000", ci)

            Case PinTypes.COUNTER, PinTypes.COUNTER_PU, _
                 PinTypes.PERIOD, PinTypes.PERIOD_PU
                If ConvertToFreq Then
                    Return PinTypeToString(PinType) & vbCrLf & _
                            "--------------" & vbCrLf & _
                            "  Raw: " & Value_RawUinteger.ToString("00000") & vbCrLf & _
                            " Freq: " & Freq.ToString("00000") & vbCrLf & _
                            " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                            " Slot: " & Value.ToString("0.00", ci)
                Else
                    Return PinTypeToString(PinType) & vbCrLf & _
                            "--------------" & vbCrLf & _
                            "  Raw: " & Value_RawUinteger.ToString("00000") & vbCrLf & _
                            " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                            " Slot: " & Value.ToString("00000")
                End If

            Case PinTypes.STEPPER_DIR
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        "  Raw: " & Value_RawInteger.ToString("00000") & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & vbCrLf & _
                        " Error" & vbCrLf & " (mm): " & Value.ToString("0.00", ci)

            Case PinTypes.GEN_IN_8, PinTypes.GEN_IN_16, PinTypes.GEN_IN_24
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        "  Raw: " & Value_RawUinteger.ToString("00000") & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                        " Slot: " & Value.ToString("00000")

            Case PinTypes.GEN_IN_FLOAT
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                        " Slot: " & Value.ToString("0.00", ci)

            Case PinTypes.ENCODER_A, PinTypes.ENCODER_A_PU
                Return PinTypeToString(PinType) & vbCrLf & _
                        "--------------" & vbCrLf & _
                        "  Raw: " & Value_RawUinteger.ToString("00000") & vbCrLf & _
                        " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                        " Slot: " & Value.ToString("00000")

                'Case PinTypes.ENCODER_B, PinTypes.ENCODER_B_PU
                '    Return PinTypeToString(PinType) & vbCrLf & _
                '            "--------------" & vbCrLf & _
                '            "  Raw: " & Value_RawUinteger.ToString("00000") & vbCrLf & _
                '            " Norm: " & Value_Normalized.ToString("0.00", ci) & vbCrLf & _
                '            " Slot: " & Value.ToString("00000")

                ' --------------------------------------------------------------------- 
            Case Else
                Return PinTypeToString(PinType)

        End Select
        Return ""
    End Function

End Class


